import ast

import pytest

from pytojsonschema.common import init_typing_namespace, init_schema_map, get_ast_name_or_attribute_string


def test_init_typing_namespace():
    assert init_typing_namespace() == {
        "Union": set(),
        "List": set(),
        "Dict": set(),
        "Optional": set(),
        "Any": set(),
        "TypedDict": set(),
    }


def test_init_schema_map():
    assert init_schema_map() == init_schema_map()


@pytest.mark.parametrize(
    "ast_element, expected",
    [
        [ast.parse("a").body[0].value, "a"],
        [ast.parse("a.b").body[0].value, "a.b"],
        [ast.parse("a.b.c").body[0].value, "a.b.c"],
        [ast.parse("a.b.c.d").body[0].value, "a.b.c.d"],
    ],
    ids=["single", "double", "triple", "quadruple"],
)
def test_get_ast_attribute_string(ast_element, expected):
    assert get_ast_name_or_attribute_string(ast_element) == expected



import ast
import copy
import typing

BASE_SCHEMA_MAP = {
    "bool": {"type": "boolean"},
    "int": {"type": "integer"},
    "float": {"type": "number"},
    "str": {"type": "string"},
}
VALID_AST_SUBSCRIPTS = frozenset({"Union", "List", "Dict", "Optional"})
VALID_TYPES = VALID_AST_SUBSCRIPTS | frozenset({"TypedDict", "Any"})

TypingNamespace = typing.Dict[str, typing.Set[str]]
Schema = typing.Dict[str, typing.Any]
SchemaMap = typing.Dict[str, Schema]


class InvalidTypeAnnotation(Exception):
    pass


def init_typing_namespace() -> TypingNamespace:
    """
    Initialize the typing namespace, whose keys are the supported typing types and the values how we can use them.
    Initial values are empty sets for all the typing types we support (keys)

    :return: A TypingNamespace object
    """
    return {valid_type: set() for valid_type in VALID_TYPES}


def init_schema_map() -> SchemaMap:
    """
    Initialize the schema map, a dictionary containing the types that translate to a concrete json schema.
    Initial values are the schemas for the bool, int, float, str and None types (keys).

    :return: A SchemaMap object
    """
    return copy.deepcopy(BASE_SCHEMA_MAP)


def get_ast_name_or_attribute_string(ast_element: typing.Union[ast.Name, ast.Attribute]) -> str:
    """
    Get the string representation of an ast name or ast attribute element.

    :param ast_element: An ast name or ast attribute element
    :return: A string
    """
    if isinstance(ast_element, ast.Name):
        return ast_element.id
    elif isinstance(ast_element.value, ast.Name):
        return f"{ast_element.value.id}.{ast_element.attr}"
    else:
        return f"{get_ast_name_or_attribute_string(ast_element.value)}.{ast_element.attr}"



import ast
import functools
import os
import tempfile

import pytest

from pytojsonschema.common import init_schema_map, InvalidTypeAnnotation
from pytojsonschema.functions import process_function_def, process_file, process_package

from .conftest import assert_expected, TEST_TYPING_NAMESPACE


@pytest.mark.parametrize(
    "ast_function_def, typing_namespace, schema_map, expected",
    [
        [
            ast.parse("def foo(a, /): pass").body[0],
            TEST_TYPING_NAMESPACE,
            init_schema_map(),
            InvalidTypeAnnotation("Function 'foo' contains positional only arguments"),
        ],
        [
            ast.parse("def foo(a, *args): pass").body[0],
            TEST_TYPING_NAMESPACE,
            init_schema_map(),
            InvalidTypeAnnotation("Function 'foo' contains a variable number positional arguments i.e. *args"),
        ],
        [
            ast.parse("def foo(**bar): pass").body[0],
            TEST_TYPING_NAMESPACE,
            init_schema_map(),
            InvalidTypeAnnotation("Function 'foo' is missing its **bar type annotation"),
        ],
        [
            ast.parse("def foo(**bar: int): pass").body[0],
            TEST_TYPING_NAMESPACE,
            init_schema_map(),
            {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "properties": {},
                "required": [],
                "additionalProperties": {"type": "integer"},
            },
        ],
        [
            ast.parse("def foo(a): pass").body[0],
            TEST_TYPING_NAMESPACE,
            init_schema_map(),
            InvalidTypeAnnotation("Function 'foo' is missing type annotation for the parameter 'a'"),
        ],
        [
            ast.parse("def foo(a: int = 3): pass").body[0],
            TEST_TYPING_NAMESPACE,
            init_schema_map(),
            {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "properties": {"a": {"type": "integer"}},
                "required": [],
                "additionalProperties": False,
            },
        ],
        [
            ast.parse("def foo(a: int): pass").body[0],
            TEST_TYPING_NAMESPACE,
            init_schema_map(),
            {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "properties": {"a": {"type": "integer"}},
                "required": ["a"],
                "additionalProperties": False,
            },
        ],
    ],
    ids=[
        "posonly_args",
        "args",
        "missing_kwargs_annotation",
        "valid_kwargs",
        "missing_arg",
        "arg_default",
        "arg_no_default",
    ],
)
def test_process_function_def(ast_function_def, typing_namespace, schema_map, expected):
    assert_expected(functools.partial(process_function_def, ast_function_def, typing_namespace, schema_map), expected)


@pytest.mark.parametrize(
    "file_contents, include_patterns, exclude_patterns, expected",
    [
        [
            "import typing\n\n\ndef foo(a: int): pass\n\n\neval(3)",
            None,
            None,
            {
                "foo": {
                    "$schema": "http://json-schema.org/draft-07/schema#",
                    "type": "object",
                    "properties": {"a": {"type": "integer"}},
                    "required": ["a"],
                    "additionalProperties": False,
                }
            },
        ],
        ["import typing\n\n\ndef foo(a: int): pass\n\n\neval(3)", ["_*"], None, {}],
        ["import typing\n\n\ndef foo(a: int): pass\n\n\neval(3)", ["*"], ["foo"], {}],
    ],
    ids=["no_patterns", "include_missing", "exclude_overwrites"],
)
def test_process_file(file_contents, include_patterns, exclude_patterns, expected):
    with tempfile.NamedTemporaryFile("w") as f:
        f.write(file_contents)
        f.flush()
        assert process_file(f.name, include_patterns, exclude_patterns) == expected


def test_process_package():
    expected = {
        "example.service.start": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "additionalProperties": False,
            "properties": {
                "service": {
                    "additionalProperties": False,
                    "properties": {
                        "address": {"type": "string"},
                        "config": {
                            "additionalProperties": {
                                "anyOf": [
                                    {"type": "object"},
                                    {"type": "array"},
                                    {"type": "null"},
                                    {"type": "string"},
                                    {"type": "boolean"},
                                    {"type": "integer"},
                                    {"type": "number"},
                                ]
                            },
                            "type": "object",
                        },
                        "debug": {"type": "boolean"},
                        "port": {"anyOf": [{"type": "integer"}, {"type": "number"}]},
                        "tags": {"items": {"type": "string"}, "type": "array"},
                    },
                    "required": ["address", "port", "config", "tags", "debug"],
                    "type": "object",
                }
            },
            "required": ["service"],
            "type": "object",
        },
        "example.service._secret": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "additionalProperties": False,
            "properties": {"secret": {"anyOf": [{"type": "string"}, {"type": "null"}]}},
            "required": [],
            "type": "object",
        },
        "example.version": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "additionalProperties": False,
            "properties": {},
            "required": [],
            "type": "object",
        },
    }
    assert process_package(os.path.join("test", "example")) == expected
    current_dir = os.getcwd()
    os.chdir("test")
    try:
        assert process_package("example") == expected
    finally:
        os.chdir(current_dir)
